/*
 * main.c - main file for Milena TTS system
 * Copyright (C) Bohdan R. Rau 2008 <ethanak@polip.com>
 * 
 * Milena is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * Milena is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Milena.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/stat.h>

#include "milena.h"
#include "milena_mbrola.h"
#include "input.h"

#define PROGRAM_VERSION "0.1.8"

struct milena *milena;
struct milena_mbrola_cfg *mbrola;
int ex_flags;
int daisy;

void translate_phrase(char *phrase,char **otbuf,int *blen)
{
	static char *buf1,*buf2;
	static int len1,len2;
	int n;
	if (!len1) {
		buf1=malloc(len1=8192);
	}
	if (!len2) {
		buf2=malloc(len2=8192);
	}
	n=milena_Prestresser(milena,phrase,buf1,len1);
	if (n) {
		len1=n+1024;
		buf1=realloc(buf1,len1);
		milena_Prestresser(milena,phrase,buf1,len1);
	}
	n=milena_TranslatePhrase(milena,(unsigned char *)buf1,buf2,len2,0);
	if (n) {
		len2=n+1024;
		buf2=realloc(buf2,len2);
		milena_TranslatePhrase(milena,(unsigned char *)buf1,buf2,len2,0);
	}
	n=milena_Poststresser(buf2,*otbuf,*blen);
	if (n) {
		*blen = n+1024;
		*otbuf=realloc(*otbuf,*blen);
		
	}
	milena_Poststresser(buf2,*otbuf,*blen);
}

int get_stressed_syllable(char *str)
{
	int syl=1;
	for (;*str;str++) {
		if (*str=='~') {
			str++;
			if (!*str) return 0;
			if (*str=='!') return syl;
			continue;
		}
		if (strchr("aeiouyYMOE",*str)) syl++;
	}
	return 0;
}

int count_syllables(char *str)
{
	int syl=0;
	for (;*str;str++) {
		if (strchr("aeiouyYMOE",*str)) syl++;
	}
	return syl;
}


void translate_par(char *par,FILE *f)
{
	static char *phrase_buf,*sentence_buf,*second_buf;
	static int phrase_len,sentence_len,second_len;
	int this_ptyp,next_ptyp;
	int have_phrase=0;
	int get_phrase()
	{
		int nlen;
		char *opar=par;
		if (have_phrase) {
			char *c=sentence_buf;
			int n=sentence_len;
			sentence_buf=second_buf;
			second_buf=c;
			sentence_len=second_len;
			second_len=n;
			have_phrase=0;
			this_ptyp=next_ptyp;
			return 1;
		}
		
		nlen=milena_GetPhrase(milena,&par,phrase_buf,phrase_len,&this_ptyp);
		if (nlen<0) return 0;
		if (nlen) {
			phrase_len=nlen+1024;
			phrase_buf=realloc(phrase_buf,phrase_len);
			milena_GetPhrase(milena,&par,phrase_buf,phrase_len,&this_ptyp);
		}
		translate_phrase(phrase_buf,&sentence_buf,&sentence_len);
		if (daisy) {
			while (opar < par && isspace(*opar)) opar++;
			fprintf(f,";PHR %d %-*.*s\n",this_ptyp,par-opar,par-opar,opar);
		}
		return 1;
	}
	
	int get_next_phrase()
	{
		int nlen;
		nlen=milena_GetPhrase(milena,&par,phrase_buf,phrase_len,&next_ptyp);
		if (nlen<0) return 0;
		if (nlen) {
			phrase_len=nlen+1024;
			phrase_buf=realloc(phrase_buf,phrase_len);
			milena_GetPhrase(milena,&par,phrase_buf,phrase_len,&next_ptyp);
		}
		translate_phrase(phrase_buf,&second_buf,&second_len);
		have_phrase=1;
		return 1;
		
	}
	if (!phrase_len) {
		phrase_buf=malloc(phrase_len=8192);
		sentence_buf=malloc(sentence_len=8192);
		second_buf=malloc(second_len=8192);
	}
	for (;;) {
		if (!get_phrase()) return;
		if (!daisy && (this_ptyp & 7) == 1) {
			if (!get_next_phrase()) this_ptyp=(this_ptyp & 0xf8) | 5;
			else {
				//int csy=get_stressed_syllable(second_buf);
				int b=0;
				if (next_ptyp == 2) {
					if (count_syllables(second_buf)<6) b=1;
				}
				//if (!b && csy>3) b=1;
				if (b) this_ptyp=(this_ptyp & 0xf8) | 4;
			}
			
		}
		milena_ModMbrolaGenPhrase(mbrola,sentence_buf,f,this_ptyp | ex_flags);
	}
}

static char *DESCR_STRING="Milena - translator tekstu polskiego na fonemy Mbroli\n\
Wersja: "PROGRAM_VERSION"\n\
Wersja modulu translacji: "MILENA_VERSION"\n\
Wersja modulu Mbroli: "MILENA_MBR_VERSION"\n\
(C) 2008 Bohden R. Rau <ethanak@polip.com>\n\n";

void help(char *name)
{
	fprintf(stderr,DESCR_STRING);
	fprintf(stderr,"Sposob uzycia:\n\t%s [opcje]\n\n",name);
	fprintf(stderr," -h	help\n");
	fprintf(stderr," -v	verbose, pokazuje etapy posrednie\n	i zastosowane reguly translacji\n	Wylacza generowalie fonemow mbroli");
	fprintf(stderr," -p	phraser, wylacznie etap frazera\n	do testow regul rozpoznawania i slownika\n");
	fprintf(stderr," -l	lektor, uwzglednia dialogi i puste akapity\n");
	fprintf(stderr," -d	Uwzglednia dodatkowe informacje dla daisy. Implikuje -l\n");
	fprintf(stderr," -i	frazer ignoruje informacje dodatkowe w {} i []\n");
	fprintf(stderr," -I	frazer ignoruje akcenty i separatory\n");
	fprintf(stderr," -E	odczytuje akapity nie zawierajace liter ani cyfr\n");
	fprintf(stderr," -e	odczytuje emoticony\n");
	fprintf(stderr," -P	frazer odczytuje wiekszosc znakow nieliterowych\n\t(implikuje -i -I -E)\n");
	fprintf(stderr," -H	frazer odczytuje godziny\n");
	fprintf(stderr," -L xx	frazer wczytuje plik jezykowy\n");
	fprintf(stderr," -u	wczytuje dodatkowy plik slownika\n");
	fprintf(stderr," -f	wczytuje dodatkowy plik frazera\n");
	fprintf(stderr," -t	wczytuje temat\n");
	fprintf(stderr," -m	wczytuje dodatkowy plik modulu Mbroli\n");
	fprintf(stderr," -Q	korekta dlugosci samoglosek\n");
	fprintf(stderr," -s	uwydatnienie akcentu (0..50)\n");
	fprintf(stderr,"Opcje -u, -m, -L, -t oraz -f moga wystepowac wielokrotnie.\n\n");
	exit(1);
	
}

void read_home_udic(void)
{
	char path[256];
	struct stat sb;
	sprintf(path,"%s/.milena_pl_userdic.dat",getenv("HOME"));
	if (stat(path,&sb)) return;
	milena_ReadUserDicWithFlags(milena,path,MILENA_UDIC_IGNORE);
	
}

main(int argc,char *argv[])
{
	char obuf[8192];
	char ibuf[8192];
	char ixbuf[8192];
	char *psf;
	int ptype,debug,phronly=0,phmode=0,stresext=-1;
	int bookmode=0,partyp,lastpar=-1,pau,spell_empty=0;
	void *inputctl;
	int got_home_udic=0;
	
	debug=0;
	
	milena=milena_Init(
		milena_FilePath("pl_pho.dat",obuf),
		milena_FilePath("pl_dict.dat",ibuf),
		milena_FilePath("pl_stress.dat",ixbuf));
	milena_ReadPhraser(milena,
		milena_FilePath("pl_phraser.dat",ibuf));
	mbrola=milena_InitModMbrola(
		milena_FilePath("pl_mbrola.dat",ibuf));
	milena_ReadUserDic(milena,
		milena_FilePath("pl_udict.dat",ibuf));
	milena_ReadInton(mbrola,
		milena_FilePath("pl_intona.dat",ibuf));
	for(;;) {
		int c=getopt(argc,argv,"vVu:f:ldphiIm:PHL:EeQs:t:");
		if (c<0) break;
		if (c=='v') {
			debug|=1;
			continue;
		}
		if (c=='V') {
			debug|=2;
			continue;
		}
		if (c=='E') {
			spell_empty=1;
			continue;
		}
		if (c=='Q') {
			milena_ModMbrolaSetFlag(
				mbrola,-1,MILENA_MBFL_EQUALIZE);
			continue;
		}
		if (c=='i') {
			phmode |= MILENA_PHR_IGNORE_INFO;
			continue;
		}
		if (c=='I') {
			phmode |= MILENA_PHR_IGNORE_TILDE;
			continue;
		}
		
		if (c=='u') {
			if (!got_home_udic) {
				read_home_udic();
				got_home_udic=1;
			}
			milena_ReadUserDicWithFlags(milena,optarg,MILENA_UDIC_IGNORE);
			continue;
		}
		if (c=='f') {
			if (!got_home_udic) {
				read_home_udic();
				got_home_udic=1;
			}
			milena_ReadPhraser(milena,optarg);
			continue;
		}
		if (c=='m') {
			milena_ModMbrolaExtras(mbrola,optarg);
			continue;
		}
		if (c=='l') {
			bookmode=1;
			continue;
		}
		if (c=='d') {
			bookmode=daisy=1;
			continue;
		}
		if (c=='p') {
			phronly=1;
			continue;
		}
		if (c=='P') {
			phmode|=4;
			spell_empty=1;
			continue;
		}
		if (c=='H') {
			milena_ReadPhraser(milena,
				milena_FilePath("pl_hours.dat",ibuf));
			continue;
		}
		if (c=='e') {
			milena_ReadPhraser(milena,
				milena_FilePath("pl_emots.dat",ibuf));
			continue;
		}
		if (c=='s') {
			stresext=strtol(optarg,NULL,10);
			continue;
		}
		if (c=='L') {
			sprintf(obuf,"pl_%s_udic.dat",optarg);
			milena_ReadUserDic(milena,
				milena_FilePath(obuf,ibuf));
			sprintf(obuf,"pl_%s_stress.dat",optarg);
			milena_ReadStressFile(milena,
				milena_FilePath(obuf,ibuf));
			milena_SetLangMode(milena,optarg);
			continue;
		}
		if (c=='t') {
			sprintf(obuf,"pl_%s_theme.dat",optarg);
			milena_ReadPhraser(milena,
				milena_FilePath(obuf,ibuf));
			continue;
		}
		help(argv[0]);
	}
	if (!got_home_udic) {
		read_home_udic();
		got_home_udic=1;
	}
	if (phmode & 4) {
		milena_ReadPhraser(milena,milena_FilePath("pl_charspell.dat",ibuf));
	}
	else if (phmode & 3) milena_SetPhraserMode(milena,phmode & 3);
	if (stresext>=0) {
		milena_ModMbrolaSetVoice(mbrola,
			MILENA_MBP_STRESLEN,stresext);
	}
	ex_flags = 0;
	if (bookmode) {
		ex_flags|=32;
		if (milena_ReadVerbs(milena,
				milena_FilePath("pl_verbs.dat",ibuf))) ex_flags|=64;;
		
	}
	inputctl=alloc_input_buffer();
	for (;;) {
		psf=fgets_ra(inputctl,stdin);
		if (!psf) break;
		partyp=milena_GetParType(milena,&psf,spell_empty);
		if (!partyp) {
			if (lastpar>0) lastpar=0;
			continue;
		}
		if (lastpar<0) {
			pau=0;
		}
		else {
			if (!bookmode) pau=MILENA_MBRK_NORMAL;
			else {
				if (lastpar==0) pau=MILENA_MBRK_LONG;
				else if (lastpar == partyp) pau=MILENA_MBRK_NORMAL;
				else if (partyp == MILENA_PARTYPE_DIALOG) pau=MILENA_MBRK_PREDIAL;
				else pau=MILENA_MBRK_POSTDIAL;
			}
		}
		lastpar=partyp;
		if (pau && !phronly) {
			milena_ModMbrolaBreak(mbrola,stdout,pau);
			if (daisy) printf(";PAR %d\n",partyp);
		}
		if (debug || !bookmode) for (;;) {
			int n;
			n=milena_GetPhrase(milena,&psf,ibuf,8192,&ptype);
			if (n) break;
			if (debug || phronly) fprintf(stderr,"%d: %s\n",ptype,ibuf);
			if (phronly) continue;
			milena_Prestresser(milena,ibuf,obuf,8192);
			if (debug) fprintf(stderr,"%s\n",obuf);
			milena_TranslatePhrase(milena,(unsigned char *)obuf,ibuf,8192,debug);
			if (debug) fprintf(stderr,"%s\n",ibuf);
			milena_Poststresser(ibuf,obuf,8192);
			if (debug) fprintf(stderr,"%s\n",obuf);
			if (!(debug & 2)) milena_ModMbrolaGenPhrase(mbrola,obuf,stdout,ptype | ex_flags);
		}
		else	translate_par(psf,stdout);
	}
	
}
